/*
hopen.c Version 0.2.0  Hash open database
Copyright (C) 2003-2010  dondalah721@yahoo.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

#include <bthash.h>

#define HOPEN 200

hashfmt *hopen(dbname,rwflag,lock,reason)
char *dbname;
int rwflag;
int lock;
int *reason;
   {
   int rslt;
   int namelen;
   int rdlen;
   int hndl;
   unsigned char *p,*q;
   hashfmt *hshhdr;
   namelen =  strlen(dbname);
   if (namelen < 1 || namelen > 128)
      {
      fprintf(stderr,"hopen: wrong "
	 "length dbname %s\n",
	 dbname);
      *reason = HASH_ERR + HOPEN + 9;
      return(NULL);
      } /* dbname wrong length */
   if (rwflag == UPDT)
      hndl = open(dbname,O_RDWR);
   else if (rwflag == READ)
      hndl = open(dbname,O_RDONLY);
   else
      {
      fprintf(stderr,"hopen: invalid read/write flag\n");
      *reason = HASH_ERR + HOPEN + 1;
      return(NULL);
      } /* rwflag err */
   if (hndl < 0)
      {
      fprintf(stderr,"hopen: %s: open ", dbname);
      if (rwflag == READ)
         fprintf(stderr,"read-only\n");
      else if (rwflag == UPDT)
         fprintf(stderr,"read/write\n");
      perror("Open error");
      *reason = SYS_ERR + errno;
      return(NULL);
      } /* if open error */
   if (lock == LOCK)
      {
      rslt = lockf(hndl,F_TLOCK,0);
      if (rslt < 0)
         {
         fprintf(stderr,"hopen: lock "
	    "error on %s\n", dbname);
         perror("Lock error");
         rslt = close(hndl);
         *reason = SYS_ERR + errno;
         return(NULL);
         } /* if lock failed */
      } /* if lock */
   else if (lock == TSTLOCK)
      {
      rslt = lockf(hndl,F_TEST,0);
      if (rslt < 0)
         {
         fprintf(stderr,"hopen: lock test "
	    "error on %s\n", dbname);
         perror("Lock test error");
         rslt = close(hndl);
         *reason = SYS_ERR + errno;
         return(NULL);
         } /* if lock test failed */
      } /* if lock test */
   else if (lock == NOLOCK);
   else
      {
      fprintf(stderr,"hopen: invalid lock command\n");
      rslt = close(hndl);
      *reason = HASH_ERR + HOPEN + 2;
      return(NULL);
      } /* lock cmd error */
   hshhdr = (hashfmt *) malloc(HDRLEN+HDRLEN);
   if (hshhdr == NULL)
      {
      fprintf(stderr,"hopen: out of memory "
	 "allocating hshhdr\n");
      rslt = lockf(hndl,F_ULOCK,0);
      rslt = close(hndl);
      *reason = HASH_ERR + HOPEN + 3;
      return(NULL);
      } /* out of mem */
   rdlen = read(hndl,hshhdr,HDRLEN);
   if (rdlen != HDRLEN)
      {
      fprintf(stderr,"hopen: dbname %s\n", dbname);
      perror("hopen: error reading header");
      rslt = lockf(hndl,F_ULOCK,0);
      rslt = close(hndl);
      free(hshhdr);
      *reason = HASH_ERR + HOPEN + 4;
      return(NULL);
      } /* read err */
   if (strncmp(hshhdr->signature,
      H_SIGNATURE,sizeof(H_SIGNATURE)))
      {
      fprintf(stderr,"hopen: dbname %s\n", dbname);
      fprintf(stderr,"hopen: error "
	 "opening non-hash file\n");
      rslt = lockf(hndl,F_ULOCK,0);
      rslt = close(hndl);
      free(hshhdr);
      *reason = HASH_ERR + HOPEN + 5;
      return(NULL);
      } /* invalid signature */
   p = (unsigned char *) hshhdr + HDRLEN;
   q = (unsigned char *) p + HDRLEN;
   while (p < q) *p++ = '\0';
   hshhdr->hndl = hndl;
   hshhdr->lock = lock;

   hshhdr->rcdnode = hshhdr->rcdhdr + hshhdr->rcdlen;

   { /* calculate offset of freestk */
   int diff;
   hashfmt tmphdr;
   p = (unsigned char *) &tmphdr.freestk;
   q = (unsigned char *) &tmphdr.signature[0];
   diff = (unsigned int) p - (unsigned int) q;
   hshhdr->fsofst  = (off_t) diff;
   } /* calculate offset of freestk */

   hshhdr->hashend = (off_t) HDRLEN
      + (hshhdr->totrcds * hshhdr->rcdnode);

   hshhdr->node = (h_nodefmt *) malloc(hshhdr->rcdnode+16);
   if (hshhdr->node == NULL)
      {
      fprintf(stderr,"hopen: out of memory "
	 "allocating hshhdr->buf1\n");
      rslt = lockf(hndl,F_ULOCK,0);
      hclose(hshhdr);
      free(hshhdr);
      *reason = HASH_ERR + HOPEN + 6;
      return(NULL);
      } /* out of mem */
   hshhdr->rcd = (unsigned char *)
      hshhdr->node + hshhdr->rcdhdr;
   hshhdr->key = (unsigned char *)
      hshhdr->rcd + hshhdr->keyofst;
   hshhdr->nodenext = (h_nodefmt *) malloc(hshhdr->rcdnode+16);
   if (hshhdr->nodenext == NULL)
      {
      fprintf(stderr,"hopen: out of memory "
	 "allocating hshhdr->buf2\n");
      rslt = lockf(hndl,F_ULOCK,0);
      hclose(hshhdr);
      free(hshhdr->node);
      free(hshhdr);
      *reason = HASH_ERR + HOPEN + 7;
      return(NULL);
      } /* out of mem */
   p = (unsigned char *) hshhdr->nodenext;
   q = (unsigned char *) p + hshhdr->rcdnode;
   while (p < q) *p++ = '\0';
   hshhdr->nodeprev = (h_nodefmt *) malloc(hshhdr->rcdnode+16);
   if (hshhdr->nodeprev == NULL)
      {
      fprintf(stderr,"hopen: out of memory "
	 "allocating hshhdr->buf3\n");
      rslt = lockf(hndl,F_ULOCK,0);
      hclose(hshhdr);
      free(hshhdr->node);
      free(hshhdr->nodenext);
      free(hshhdr);
      *reason = HASH_ERR + HOPEN + 8;
      return(NULL);
      } /* out of mem */
   p = (unsigned char *) hshhdr->nodeprev;
   q = (unsigned char *) p + hshhdr->rcdnode;
   while (p < q) *p++ = '\0';
   strcpy(hshhdr->dbname,dbname);
   return(hshhdr);
   } /* hopen */
